import os
import sys
from .functions import definitions
from .expression import add_context_processor
from .builder import build_all
from .dumper import dump
from .tree import NodeException


for definition in definitions:
    add_context_processor(definition)

class ErrorList(list):
    def __init__(self):
        self.nodes = set()
    def append(self, error):
        if not isinstance(error, NodeException) or error.node not in self.nodes:
            super().append(error)
        if isinstance(error, NodeException):
            self.nodes.add(error.node)

def process(input, output, filename=None, errors=[]):
    if filename:
        filename = os.path.abspath(filename)
    first = True

    for doc in build_all(input, filename, errors):
        if not first:
            output.write("---\n")
        else:
            first = False
        dump(doc, output, errors)



''' Transforms the parsed YAML structure into a tree '''

from . import loader
from .tree import (Document, Scalar, Sequence,
                   Mapping, MergeKey, NodeException)
from .expression import (Expression, Function, FunctionBlock)
from ruamel.yaml.constructor import ConstructorError
from ruamel.yaml.nodes import (ScalarNode, SequenceNode, MappingNode)
from ruamel.yaml.comments import TaggedScalar


def build(source, filepath=None, errors=None):
    yaml_doc = loader.load(source)
    return create_tree(yaml_doc, filepath, errors)


def build_all(source, filepath=None, errors=[]):
    for yaml_doc in loader.load_all(source, errors):
        yield create_tree(yaml_doc, filepath, errors)


def create_tree(yaml_doc, filepath, errors=None):
    if yaml_doc is None:
        return None
    doc = Document()
    tree = convert(yaml_doc, doc)
    doc.filepath = filepath
    doc.root = tree
    doc.errors = errors
    return tree


def convert(node, doc):
    if isinstance(node, dict):
        return Mapping([(convert(k, doc), convert(v, doc)) for (k, v) in node.items()], doc)
    if isinstance(node, list):
        return Sequence([convert(v, doc) for v in node], doc)
    if isinstance(node, TaggedNode):
        return node.create(doc)
    return Scalar(node, doc)


def construct_node(self, node):
    if isinstance(node, ScalarNode):
        return self.construct_scalar(node)
    if isinstance(node, MappingNode):
        return [c for c in self.construct_yaml_map(node)][0]
    if isinstance(node, SequenceNode):
        result = [c for c in self.construct_yaml_seq(node)][0]
        return result


class TaggedNode:
    def __init__(self, value, source=None):
        self.value = value
        self.source = source


class InvisibleNode(TaggedNode):
    def create(self, doc):
        if isinstance(self.value, str):
            result = Expression(self.value, doc, self.source)
        else:
            result = convert(self.value, doc)
        result.visible = False
        return result


def invisible_constructor(self, node):
    return InvisibleNode(construct_node(self, node), node)


loader.add_constructor('!-', invisible_constructor)


class ExpressionNode(TaggedNode):
    def create(self, doc):
        result = Expression(self.value, doc, self.source)
        return result


def expression_constructor(self, node):
    if isinstance(node, ScalarNode):
        return ExpressionNode(self.construct_scalar(node), node)
    else:
        raise ConstructorError(
            None, None,
            'expected a scalar node, but found %s' % node.id,
            node.start_mark)


loader.add_constructor('!?', expression_constructor)


class FunctionNode(TaggedNode):
    def create(self, doc):
        result = Function(self.value, doc, self.source)
        return result


class FunctionBlockNode(TaggedNode):
    def __init__(self, value, constructor):
        self.value = value
        self.constructor = constructor
        self.node = construct_node(self.constructor, self.value)

    def create(self, doc):
        result = FunctionBlock(lambda: convert(self.node, doc))
        return result


def function_constructor(self, node):
    if isinstance(node, ScalarNode):
        return FunctionNode(self.construct_scalar(node))
    else:
        return FunctionBlockNode(node, self)


loader.add_constructor('!()', function_constructor)


class IfNode(TaggedNode, MergeKey):
    def __init__(self, doc, node):
        TaggedNode.__init__(self, doc, node)
        MergeKey.__init__(self)

    def create(self, doc):
        self.expression = Expression(self.value, doc, self.source)
        return self

    def merge(self, parent, source):
        if self.expression.content():
            MergeKey.merge(self, parent, source)


def if_constructor(self, node):
    if isinstance(node, ScalarNode):
        return IfNode(self.construct_scalar(node), node)
    assert False, 'expression nodes must be scalar'


loader.add_constructor('!if', if_constructor)


class ForNode(TaggedNode, MergeKey):
    def __init__(self, doc, node):
        TaggedNode.__init__(self, doc, node)
        MergeKey.__init__(self)

    def create(self, doc):
        self.expression = Expression(self.value, doc, self.source)
        return self

    def merge(self, parent, source):
        try:
            for item in self.expression.content():
                MergeKey.merge(self, parent, source.content()(item))
        except Exception as e:
            raise NodeException(self.expression, e)


def for_constructor(self, node):
    if isinstance(node, ScalarNode):
        return ForNode(self.construct_scalar(node), node)
    assert False, 'expression nodes must be scalar'


loader.add_constructor('!for', for_constructor)



''' Model for the internal document structure. '''

from collections import OrderedDict


class Document:
    def __init__(self):
        self.filepath = None
        self.root = None
        self.errors = None
        

class Mark:
    def __init__(self, line, column):
        self.line = line
        self.column = column


class NodeException(Exception):
    def __init__(self, node, cause):
        self.node = node
        self.cause = cause

    def __str__(self):
        filepath = self.node.doc.filepath
        if filepath is None:
            filepath = '<stdin>'
        start_mark = self.start_mark()
        line = start_mark.line
        column = start_mark.column

        if isinstance(self.cause, KeyError):
            message = 'key not found: '+str(self.cause)
        else:
            message = str(self.cause)

        if not isinstance(self.cause, NodeException):
            message = message+'\n in "'+filepath + \
                '", line ' + str(start_mark.line+1) + \
                ', column ' + str(start_mark.column+1) + \
                ':\n  '+str(self.value())
        return message

    def value(self):
        if self.node.source is not None:
            return self.node.source.value

    def start_mark(self):
        if self.node.source is None:
            return Mark(0, 0)
        column = self.node.source.end_mark.column - len(self.node.source.value)
        if hasattr(self.cause, 'position') and self.cause.position:
            column += self.cause.position
            if self.node.source.value.startswith('.'):
                column -= 2
        return Mark(self.node.source.start_mark.line, column)

    def end_mark(self):
        if self.node.source is None:
            return Mark(0, 0)
        return self.node.source.end_mark


class Node:
    def __init__(self, doc, source=None):
        self.visible = True
        self.memo = None
        self.doc = doc
        self.source = source

    def content(self):
        if(self.memo is None):
            self.memo = self.create_content()
        return self.memo

    def create_content(self):
        return self

    def receive(self, visitor):
        visitor(self.content())


class MergeKey:
    def __init__(self):
        self.visited = False

    def merge(self, parent, source):
        self.visited = True    
        if isinstance(parent, Mapping):
            if isinstance(source, OrderedDict):
                parent.update(OrderedDict.items(source))
            elif isinstance(source, dict):
                parent.update(source)
            elif isinstance(source, Node):
                self.merge(parent, source.content())
            if isinstance(source, Mapping):
                parent.special_entries.extend(source.special_entries)
        else:
            parent.extend(source)


class Scalar(Node):
    def __init__(self, value, doc=None):
        Node.__init__(self, doc)
        self.memo = value

    def content(self):
        return self.memo

    def __repr__(self):
        return str(self.memo)


class Mapping(OrderedDict, Node):
    def __init__(self):
        Node.__init__(self, None)

    def __init__(self, source=None, doc=None):
        Node.__init__(self, doc)
        self.special_entries = []
        if source:
            if isinstance(source, OrderedDict):
                source = OrderedDict.items(source)
            elif isinstance(source, dict):
                source = dict.items(source)
            OrderedDict.__init__(self, self.handle_keys(source))

    def __getitem__(self, key):
        if(self.__contains__(key)):
            return super().__getitem__(key).content()
        else:
            return self.resolve_special(key)

    def items(self):
        self.resolve_special()
        for (k, v) in super().items():
            if isinstance(v, Node):
                result = (k, v.content())
                if v.visible:
                    yield result
            elif v is not None:
                yield (k, v)

    def __eq__(self, other):
        return dict(self.items()) == other

    def __repr__(self):
        return '<mapping>'

    def receive(self, visitor):
        for (k, v) in self.items():
            if isinstance(v, Node):
                v.receive(visitor)
            else:
                visitor(v)

    def handle_keys(self, items):
        for (k, v) in items:
            if isinstance(k, Scalar):
                yield (k.content(), v)
            else:
                self.special_entries.insert(0, (k, v))

    def resolve_special(self, key=None):
        while len(self.special_entries) > 0:
            (k, v) = self.special_entries.pop()
            if isinstance(k, MergeKey):
                k.merge(self, v)
            else:
                computed_key = k.content()
                if computed_key is not None:
                    self[computed_key] = v
                else:
                    raise NodeException(k, "mapping key is null")
        self.special_entries = []
        if key is not None:
            return super().__getitem__(key)


class Sequence(list, Node):
    def __init__(self, value=None, doc=None):
        Node.__init__(self, doc)
        if value:
            list.__init__(self, value)

    def __getitem__(self, index):
        return super().__getitem__(index).content()

    def __eq__(self, other):
        return list(self) == other

    def __iter__(self):
        iter = super().__iter__()
        try:
            while True:
                node = iter.__next__()
                if isinstance(node, Mapping) and len(node) == 0 and len(node.special_entries) == 1:
                    (k, v) = node.special_entries[0]
                    if not k.visited:
                        k.merge(self, v)
                elif isinstance(node, Node):
                    if node.content() is not None and node.visible:
                        yield node.content()
                else:
                    yield node
        except StopIteration:
            return

    def __repr__(self):
        return '<sequence>'

    def receive(self, visitor):
        for v in self:
            if isinstance(v, Node):
                v.receive(visitor)
            else:
                visitor(v)



''' Serialization into YAML text '''

import ruamel.yaml
from .loader import SimpleString
from .tree import (Mapping, Sequence, Scalar, NodeException)
from io import StringIO


def dump(tree, output, errors=[]):
    yaml = ruamel.yaml.YAML()

    def represent_mapping(representer, mapping):
        filtered = []
        for (k,v) in mapping.items():
            if v is not None:
                filtered.append((k,v))
        return representer.represent_mapping(u'tag:yaml.org,2002:map', dict(filtered))

    def represent_sequence(representer, sequence):
        return representer.represent_sequence(u'tag:yaml.org,2002:seq', sequence)

    def represent_string(representer, string):
        return representer.represent_scalar(u'tag:yaml.org,2002:str', string)

    yaml.representer.add_representer(Mapping, represent_mapping)
    yaml.representer.add_representer(Sequence, represent_sequence)
    yaml.representer.add_representer(SimpleString, represent_string)

    if tree is not None:
        try:
            out = StringIO()
            yaml.dump(tree, out)
            if len(errors) == 0:
                output.write(out.getvalue())
        except ruamel.yaml.representer.RepresenterError as error:
            errors.append(NodeException(
                tree, "document is not a mapping nor a sequence"))
        except Exception as error:
            errors.append(error)



''' Built-in functions  '''

import os
from .builder import build
from .expression import Holder

base_dir = None


def import_definition(context, root):
    if hasattr(root, 'doc') and hasattr(root.doc, 'filepath') and root.doc.filepath:
        current_dir = os.path.dirname(root.doc.filepath)
    else:
        current_dir = os.getcwd()
    global base_dir
    if base_dir is None:
        base_dir = current_dir

    def import_function(filename):
        if os.path.isabs(filename):
            filepath = filename
        else:
            filepath = os.path.join(current_dir, filename)

        import_dirs = [base_dir]
        import_path = os.environ.get('YGLU_IMPORT_ALLOW')
        if import_path is not None:
            import_dirs.extend(import_path.split(os.pathsep))

        allow = False
        for dir in import_dirs:
            relative_path = os.path.relpath(filepath, start=dir)
            if not relative_path.startswith(os.pardir):
                allow = True
                break
        if not allow:
            raise Exception(
                'Import not allowed. Set $YGLU_IMPORT_ALLOW to allow import from other directories')

        with open(filepath) as file:
            if hasattr(root, 'doc'):
                errors = root.doc.errors
            else:
                errors = None
            result = build(file, filename, errors)
            return Holder(result)

    context['$import'] = import_function


definitions = [import_definition]
